<?php
/* --------------------------------------------------------------
   ShopvoteContentViewHelper.inc.php 2019-07-11
   Gambio GmbH
   http://www.gambio.de
   Copyright (c) 2019 Gambio GmbH
   Released under the GNU General Public License (Version 2)
   [http://www.gnu.org/licenses/gpl-2.0.html]
   --------------------------------------------------------------
*/

/**
 * Class ShopvoteContentViewHelper
 *
 * Bundles functionality required by various overloads of *ContentView classes.
 *
 */
class ShopvoteContentViewHelper
{
    /**
     * Returns true if Shopvote module is installed and product reviews feature is enabled.
     *
     * @return bool
     */
    public static function isShopvoteReviewsEnabled()
    {
        static $isEnabled;
        if (isset($isEnabled)) {
            return $isEnabled;
        }
        $isInstalled    = (int)gm_get_conf('MODULE_CENTER_SHOPVOTE_INSTALLED') === 1;
        $reviewsEnabled = (int)gm_get_conf('SHOPVOTE_PRODUCTREVIEWS_ENABLED') === 1;
        
        return $isInstalled && $reviewsEnabled;
    }

    public static function getAverageRatings($productIDs)
    {
        $shopId           = new NonEmptyStringType((string)gm_get_conf('SHOPVOTE_SHOP_ID'));
        $apiKey           = new NonEmptyStringType((string)gm_get_conf('SHOPVOTE_API_KEY'));
        $languageCode     = MainFactory::create('LanguageCode', new StringType((string)$_SESSION['language_code']));
        $svReviewsFactory = MainFactory::create('ShopvoteProductReviewsFactory', $shopId, $apiKey, $languageCode);
        $shopvoteRatings  = [];
        foreach ($productIDs as $productID) {
            $productsId = new IdType((int)$productID);
            /** @var \ShopvoteProductReviewsCollection $svReviewsCollection */
            $svReviewsCollection = $svReviewsFactory->getReviewsForProduct($productsId);
            $internalRating      = static::getInternalProductRatings($productsId);
            $totalCount          = $internalRating['count'] + $svReviewsCollection->getRatingCount();
            if ($totalCount > 0) {
                $totalRating = ($internalRating['rating'] * $internalRating['count']
                                + $svReviewsCollection->getRatingValue() * $svReviewsCollection->getRatingCount())
                               / ($internalRating['count'] + $svReviewsCollection->getRatingCount());
            } else {
                $totalRating = 0;
            }
            $shopvoteRatings[$productsId->asInt()] = [
                'rating'  => $totalCount > 0 ? $totalRating : -1,
                'rounded' => $totalCount > 0 ? (int)floor($totalRating) : -1,
                'count'   => $totalCount,
            ];
        }
        return $shopvoteRatings;
    }
    
    /**
     * Retrieves product ratings from shop’s native review/ratings system (from DB)
     *
     * @param \IdType $productsId
     *
     * @return array
     */
    public static function getInternalProductRatings(IdType $productsId)
    {
        $queryString = 'SELECT
							`products_id`,
							AVG(`reviews_rating`) `rating`,
							count(`products_id`) `qty`
						FROM
							`reviews`
						WHERE
							`products_id` = ?
						GROUP BY
							`products_id`';
        $db          = StaticGXCoreLoader::getDatabaseQueryBuilder();
        /** @var \CI_DB_mysqli_result $queryResult */
        $queryResult    = $db->query($queryString, [$productsId->asInt()]);
        $productRatings = [
            'rating' => 0,
            'count'  => 0,
        ];
        if ($queryResult->num_rows() === 1) {
            $row            = $queryResult->row_array();
            $productRatings = [
                'rating' => (float)$row['rating'],
                'count'  => (int)$row['qty'],
            ];
        }
        
        return $productRatings;
    }
}
