<?php
/* --------------------------------------------------------------
   ShopvoteProductReviewsFactory.inc.php 2021-01-26
   Gambio GmbH
   http://www.gambio.de
   Copyright (c) 2019 Gambio GmbH
   Released under the GNU General Public License (Version 2)
   [http://www.gnu.org/licenses/gpl-2.0.html]
   --------------------------------------------------------------
*/

use Psr\Log\LoggerAwareInterface;
use Psr\Log\LoggerInterface;

class ShopvoteProductReviewsFactory implements LoggerAwareInterface
{
    /** @var LoggerInterface */
    protected $logger;
    
    /** @var \NonEmptyStringType */
    protected $shopId;
    
    /** @var \NonEmptyStringType */
    protected $apiKey;
    
    /** @var \LanguageCode */
    protected $languageCode;

    /** @var float */
    protected $timeout;
    
    /** @var int */
    protected $maximumCacheAge;
    
    protected $requestDays;
    
    const DEFAULT_TIMEOUT           = 2.000;
    const DEFAULT_MAXIMUM_CACHE_AGE = 86400; // 86400 = 24 hours
    const DEFAULT_REQUEST_DAYS      = 365;
    
    
    public function __construct(
        NonEmptyStringType $shopId,
        NonEmptyStringType $apiKey,
        LanguageCode $languageCode = null
    ) {
        $this->logger = MainFactory::create('ShopvoteLogger');
        $this->shopId = $shopId;
        $this->apiKey = $apiKey;
        if ($languageCode === null) {
            $languageCode = MainFactory::create('LanguageCode', new StringType($_SESSION['language_code']));
        }
        $this->languageCode    = $languageCode;
        $this->timeout         = static::DEFAULT_TIMEOUT;
        $this->maximumCacheAge = static::DEFAULT_MAXIMUM_CACHE_AGE;
        $this->requestDays     = static::DEFAULT_REQUEST_DAYS;
    }
    
    
    protected function getMaximumCacheAge()
    {
        return $this->maximumCacheAge;
    }
    
    
    public function getReviewsForProduct(IdType $productsId)
    {
        $lastUpdateTimestamp = (int)gm_get_conf('SHOPVOTE_LAST_UPDATE_TIMESTAMP');
        if ($lastUpdateTimestamp === 0 || ($lastUpdateTimestamp + $this->maximumCacheAge) < time()) {
            $updateDays = $this->requestDays;
            if ($lastUpdateTimestamp > 0) {
                $daysSinceLastUpdate = ceil((time() - $lastUpdateTimestamp) / 86400);
                $updateDays = $daysSinceLastUpdate + 1;
                $this->logger->debug('Days since last update: {days} - updating cache', ['days' => $daysSinceLastUpdate]);
            } else {
                $this->logger->debug('No update recorded, getting reviews from the last {days} days', ['days' => $updateDays]);
            }
            $this->updateReviewsCache($updateDays);
        }
        
        /** @var ProductReadService $productReadService */
        $productReadService = StaticGXCoreLoader::getService('ProductRead');
        $product            = $productReadService->getProductById($productsId);
        $gtin               = $product->getEan();
        $model              = $product->getProductModel();
        $where = []; //['gtin' => $gtin, 'sku' => $model];
        if (!empty($gtin)) {
            $where['gtin'] = $gtin;
        }
        if (!empty($model)) {
            $where['sku'] = $model;
        }
        $db                 = StaticGXCoreLoader::getDatabaseQueryBuilder();
        if ($db->table_exists('shopvote_reviews')) {
            $reviews = $db->or_where($where)->get('shopvote_reviews')->result_array();
        } else {
            $reviews = [];
        }
        $reviewsCollection  = MainFactory::create('ShopvoteProductReviewsCollection', $productsId);
        $reviewsCollection->initFromReviewsArray($reviews);
        
        return $reviewsCollection;
    }
    
    
    /**
     * @param IdType $productsId
     *
     * @return bool|mixed|ShopvoteProductReviewsCollection
     *
     * @deprecated don’t use, old API
     */
    public function _getReviewsForProduct(IdType $productsId)
    {
        $dataCache = DataCache::get_instance();
        $cacheKey = 'shopvote-productreviews-' . implode('-',
                                                         [
                                                             $this->shopId->asString(),
                                                             $productsId->asInt(),
                                                             $this->languageCode->asString(),
                                                         ]
            );
        $reviewsJson = $dataCache->key_exists($cacheKey, true) ? $dataCache->get_data($cacheKey, true) : false;
        $reviewsCollection = MainFactory::create(
            'ShopvoteProductReviewsCollection',
            $productsId,
            new NonEmptyStringType('[]')
        );
        if ($reviewsJson !== false) {
            $reviewsCollection = MainFactory::create(
                'ShopvoteProductReviewsCollection',
                $productsId,
                new NonEmptyStringType($reviewsJson)
            );
            if ($reviewsCollection->getAge() > $this->getMaximumCacheAge()) {
                $reviewsJson = false;
            }
        }
        if ($reviewsJson === false) {
            $reviewsJson = $this->fetchReviewsFromShopvote($productsId);
            $reviewsCollection = MainFactory::create(
                'ShopvoteProductReviewsCollection',
                $productsId,
                new NonEmptyStringType($reviewsJson)
            );
            $dataCache->set_data($cacheKey, $reviewsCollection->asJson(), true);
        }

        return $reviewsCollection;
    }
    
    
    /**
     * @param IdType $productsId
     *
     * @return string
     *
     * @deprecated don’t use, old API
     */
    protected function fetchReviewsFromShopvote(IdType $productsId)
    {
        die('you should not be seeing this.');
        $jwt = $this->makeJsonWebToken();
        $jwtString = (string)$jwt;

        /** @var \ProductReadService $productsRead */
        $productsRead = StaticGXCoreLoader::getService('ProductRead');
        $product = $productsRead->getProductById($productsId);
        $gtin = $product->getEan();
        $productUrl = xtc_href_link(
            FILENAME_PRODUCT_INFO,
            xtc_product_link($productsId->asInt(), $product->getName($this->languageCode)) . '&no_boost=1'
        );
        
        $url = sprintf('https://api.shopvote.de/product-reviews/v1/%d/reviews', $this->shopId->asString());
        $parameters = [
            'gtin'       => $gtin,
            'productUrl' => $productUrl,
            'sku'        => $product->getProductModel(),
            'sd'         => 'true',
            'lang'       => $this->languageCode->asString(),
            'limit'      => '50', // [1..50]
            'rs'         => 'W8F',
        ];
        $url .= '?' . http_build_query($parameters, '', '&');
    
        $requestHeaders = [
            'shopvote-authorization: Bearer ' . $jwtString,
            'Accept: application/json',
            'Content-Type: application/json',
            'User-Agent: Gambio_GX',
        ];
        $request = MainFactory::create('RestRequest', 'GET', $url, null, $requestHeaders);
        $restService = MainFactory::create('RestService');
        $restService->setTimeout($this->timeout);
        try {
            $response = $restService->performRequest($request);
            if ($response->getResponseCode() === 200) {
                return $response->getResponseBody();
            }
        } catch (RestTimeoutException $e) {
        } catch (RestException $e) {
        }
    
        return '[]';
    }
    
    
    /**
     * creates JWT for authentication against Shopvote REST API endpoint
     *
     * @return JsonWebToken
     */
    protected function makeJsonWebToken()
    {
        $iat          = time();
        $iatChecksum  = strtoupper(substr(hash('sha256', $iat), 5, 3));
        $apiKeyHashed = hash('sha256', $this->apiKey->asString());
    
        $headers = MainFactory::create(
            'KeyValueCollection',
            [
                'typ' => 'JWT',
                'alg' => 'HS256',
            ]
        );
    
        $payload = MainFactory::create(
            'KeyValueCollection',
            [
                'iat'        => $iat,
                'iatcs'      => $iatChecksum,
                'shop_id'    => $this->shopId->asString(),
                'secure_key' => $apiKeyHashed,
            ]
        );
        /** @var \JsonWebToken $jwt */
        $jwt = MainFactory::create('JsonWebToken', $headers, $payload);
        $jwtSecret = $iat . $this->apiKey->asString();
        $jwtSecret = hash('sha256', $jwtSecret);
        $jwt->setSecret(new NonEmptyStringType($jwtSecret));
        $jwt->setUseRawHmacForSignature(true);
        $jwt->setIncludeSecretInSignaturePayload(false);
        return $jwt;
    }
    
    
    /**
     * fetches product reviews from Shopvote
     *
     * @param int $days
     *
     * @return string
     */
    public function fetchLatestReviewsFromShopvote(int $days)
    {
        $jwt = $this->makeJsonWebToken();
        $jwtString = (string)$jwt;
        $url = sprintf('https://api.shopvote.de/product-reviews/v1/%s/reviews?days=%d', $this->shopId->asString(), $days);
        $requestHeaders = [
            'shopvote-authorization: Bearer ' . $jwtString,
            'Accept: application/json',
            'Content-Type: application/json',
            'User-Agent: Gambio_GX',
        ];
        $request = MainFactory::create('RestRequest', 'GET', $url, null, $requestHeaders);
        $restService = MainFactory::create('RestService');
        $restService->setTimeout($this->timeout);
        try {
            $response = $restService->performRequest($request);
            if ($response->getResponseCode() === 200) {
                return $response->getResponseBody();
            }
        } catch (RestTimeoutException $e) {
            $this->logger->debug('Timeout fetching reviews');
        } catch (RestException $e) {
            $this->logger->debug('REST error fetching reviews: ' . $e->getMessage());
        }
    
        return '[]';
    }
    
    
    /**
     * Updates cached product reviews
     *
     * @param int $daysToFetch
     *
     * @throws Exception
     */
    public function updateReviewsCache($daysToFetch = 365)
    {
        $db = StaticGXCoreLoader::getDatabaseQueryBuilder();
        if ($db->table_exists('shopvote_reviews') === false) {
            $this->logger->debug('Table shopvote_reviews does not exist, aborting update of reviews cache');
            return;
        }
        $this->logger->debug('Retrieving product reviews from the last {days} days', ['days' => $daysToFetch]);
        $reviewsJson = $this->fetchLatestReviewsFromShopvote($daysToFetch);
        $reviews     = json_decode($reviewsJson, true);
        if (!empty($reviews) && json_last_error() === JSON_ERROR_NONE) {
            foreach ($reviews['reviews'] as $productReview) {
                $reviewData = [
                    'review_uid'   => $productReview['reviewUID'],
                    'sku'          => $productReview['sku'],
                    'gtin'         => $productReview['gtin'],
                    'created'      => (new \DateTime($productReview['created']))->format('Y-m-d'),
                    'author'       => $productReview['author'],
                    'rating_value' => (int)$productReview['rating_value'],
                    'review_text'  => $productReview['text'],
                ];
                $db->replace('shopvote_reviews', $reviewData);
            }
            gm_set_conf('SHOPVOTE_LAST_UPDATE_TIMESTAMP', time());
            gm_get_conf('SHOPVOTE_LAST_UPDATE_TIMESTAMP', 'ASSOC', true);
        } else {
            $this->logger->debug('No (new) reviews within the last {days} days', ['days' => $daysToFetch]);
        }
    }
    
    public function getAllReviewsFromCache()
    {
        $db = StaticGXCoreLoader::getDatabaseQueryBuilder();
        if ($db->table_exists('shopvote_reviews')) {
            $reviews = $db->get('shopvote_reviews')->result_array();
        } else {
            $reviews = [];
        }
        return $reviews;
    }
    
    public function clearCache()
    {
        $this->logger->debug('clearing reviews cache');
        $db = StaticGXCoreLoader::getDatabaseQueryBuilder();
        if ($db->table_exists('shopvote_reviews')) {
            $success = $db->truncate('shopvote_reviews');
            gm_set_conf('SHOPVOTE_LAST_UPDATE_TIMESTAMP', 0);
        } else {
            $success = false;
        }
        return $success;
    }
    
    /**
     * @param float $timeout
     */
    public function setTimeout($timeout)
    {
        $this->timeout = (float)$timeout;
    }
    
    
    /**
     * Sets a logger instance on the object
     *
     * @param LoggerInterface $logger
     *
     * @return null
     */
    public function setLogger(LoggerInterface $logger)
    {
        $this->logger = $logger;
    }
}
